#ifndef FSGUIFILEDIALOG_IS_INCLUDED
#define FSGUIFILEDIALOG_IS_INCLUDED
/* { */

#include <fsgui.h>



/*! This class provides a functionality of file dialog.
    To enable platform-specific file dialog, this class hooks Show function and calls a platform-specific file dialog function. 
	
	FsGuiCanvas is required to use this class.  It might work without FsGuiCanvas in Windows and MacOSX.  However the behavior of the class is not guaranteed in other environment and in the future versions.  
  */
class FsGuiFileDialog : public FsGuiDialog
{
private:
	mutable YSBOOL busy;
	class FsGuiFileDialog_PlatformSpecificData *param;
	static void CloseModalCallBackFunc(class FsGuiControlBase *,class FsGuiDialog *,int);

public:
	enum MODE
	{
		MODE_OPEN,
		MODE_SAVE
	};

	/*! Input: Set YSTRUE if the file dialog to allow multiple selection. */
	YSBOOL multiSelect;

	/*! Input; Set file dialog mode. */
	MODE mode;

	/*! Input: File extensions. */
	YsArray <YsWString> fileExtensionArray;

	/*! Input: Default file name. */
	YsWString defaultFileName;

	/*! Input: Title. */
	YsWString title;

	/*! Input/Output: This will be set to YSTRUE when the user made a selection.  Can be used in the polling-based application. */
	YSBOOL selectionMade;

	/*! Output: Set to YSOK if the user clicks "OK", or YSERR otherwise. */
	YSRESULT res;

	/*! Output: Selected files. */
	YsArray <YsWString> selectedFileArray;

	/*! Default constructor. */
	FsGuiFileDialog();

	/*! Default destructor. */
	~FsGuiFileDialog();

	/*! Initializes the variables. */
	void Initialize(void);

	/*! This function will call platform-specific file dialog function, and let the user select a file(s).  When the files are selected, this function will call CloseModalDialog to give control back to the parent dialog or canvas. */
	virtual void Show(const FsGuiDialog *excludeFromDrawing=NULL) const;

private:
	void AllocatePlatformSpecificData(void);
	void DeletePlatformSpecificData(void);
};

/* } */
#endif
