#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <ysclass.h>

#ifdef WIN32
#include <windows.h>
#endif

#ifndef __APPLE__
#include <GL/gl.h>
#include <GL/glu.h>
#else
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#endif

#include <fssimplewindow.h>
#include <fsgui.h>



class FsApplicationMenu : public FsGuiPopUpMenu
{
public:
	YSBOOL terminate;

	FsGuiPopUpMenuItem *fileMenu;
	FsGuiPopUpMenuItem *fileOpen,*fileSave,*fileExit;

	FsApplicationMenu();
	void MakeMenu(void);

	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *item);
};

FsApplicationMenu::FsApplicationMenu()
{
	terminate=YSFALSE;
}

void FsApplicationMenu::MakeMenu(void)
{
	this->SetIsPullDownMenu(YSTRUE);

	fileMenu=AddTextItem(0,FSKEY_F,"File (F)");
	fileOpen=fileMenu->GetSubMenu()->AddTextItem(0,FSKEY_O,"Open (O)");
	fileSave=fileMenu->GetSubMenu()->AddTextItem(0,FSKEY_S,"Save (S)");
	fileExit=fileMenu->GetSubMenu()->AddTextItem(0,FSKEY_X,"Exit (X)");
}

void FsApplicationMenu::OnSelectMenuItem(FsGuiPopUpMenuItem *item)
{
	if(item==fileExit)
	{
		terminate=YSTRUE;
	}
}



class FsApplicationDialog : public FsGuiDialog
{
public:
	YSBOOL terminate;

	FsGuiButton *okBtn;
	FsGuiButton *cancelBtn;

	FsApplicationDialog();
	void MakeDialog(void);
	virtual void OnButtonClick(FsGuiButton *btn);
};

FsApplicationDialog::FsApplicationDialog()
{
	terminate=YSFALSE;
}

void FsApplicationDialog::MakeDialog(void)
{
	SetTopLeftCorner(0,32);

	AddStaticText(0,FSKEY_NULL,"Sample Dialog",16,1,YSTRUE);
	okBtn=AddTextButton(0,FSKEY_NULL,FSGUI_PUSHBUTTON,"OK",YSTRUE);
	cancelBtn=AddTextButton(0,FSKEY_NULL,FSGUI_PUSHBUTTON,"Cancel",YSTRUE);

	SetTransparency(YSFALSE);

	Fit();
}

void FsApplicationDialog::OnButtonClick(FsGuiButton *btn)
{
	if(btn==okBtn || btn==cancelBtn)
	{
		terminate=YSTRUE;
	}
}


static void MainLoopOneStep(void);
static FsApplicationMenu *mainMenu=NULL;
static FsApplicationDialog *mainDlg=NULL;
static const int sleepTime=20;


int main(void)
{
	srand((unsigned int)time(NULL));

	FsOpenWindow(16,16,1024,768,1);



	FsGuiDefaultAsciiFixedFontRenderer fixedFontRenderer;
	FsGuiObject::defAsciiRenderer=&fixedFontRenderer;
	FsGuiObject::defUnicodeRenderer=&fixedFontRenderer;


	mainMenu=new FsApplicationMenu;
	mainMenu->MakeMenu();

	mainDlg=new FsApplicationDialog;
	mainDlg->MakeDialog();



	FsPollDevice();
	while(YSTRUE!=mainMenu->terminate && YSTRUE!=mainDlg->terminate)
	{
		MainLoopOneStep();
	}

	return 0;
}

static void MainLoopOneStep(void)
{
	FsPollDevice();


	int wid,hei;
	FsGetWindowSize(wid,hei);



	const YSBOOL shift=(YSBOOL)FsGetKeyState(FSKEY_SHIFT);
	const YSBOOL ctrl=(YSBOOL)FsGetKeyState(FSKEY_CTRL);
	const YSBOOL alt=(YSBOOL)FsGetKeyState(FSKEY_ALT);

	int key;
	while(key=FsInkey())
	{
		mainMenu->KeyIn(key,shift,ctrl,alt);
		mainDlg->KeyIn(key,shift,ctrl,alt);
	}
	while(key=FsInkeyChar())
	{
		mainDlg->CharIn(key);
	}

	int lb,mb,rb,mx,my;
	int mouseEvent;
	while(FSMOUSEEVENT_NONE!=(mouseEvent=FsGetMouseEvent(lb,mb,rb,mx,my)))
	{
		mainMenu->SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my);
		mainDlg->SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my);
	}



	if(0!=FsCheckWindowExposure() ||
	   YSTRUE==mainMenu->NeedRedraw() ||
	   YSTRUE==mainDlg->NeedRedraw())
	{
		glClearColor(1.0,1.0,1.0,1.0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

		mainDlg->Show();
		mainMenu->Show();

		glFlush();
		FsSwapBuffers();
	}

	FsSleep(sleepTime);
}
