#ifndef FSGUIBASE_IS_INCLUDED
#define FSGUIBASE_IS_INCLUDED
/* { */



#include <ysclass.h>
#include <ysbitmap.h>
#include <ysbitmaparray.h>



class FsGuiBitmapType : public YsBitmapArray
{
public:
	int GetWidth(void) const;
	int GetHeight(void) const;
	void MakeSmallEmptyBitmap(void);
	void SetBitmap(const YsBitmap &bmpIn);
};



////////////////////////////////////////////////////////////

class FsGuiObject
{
public:
	enum SCHEME
	{
		PLAIN,
		MODERN
	};

public:
	enum FSGUI_DRAWINGTYPE
	{
		FSGUI_DRAWINGTYPE_UNKNOWN,
		FSGUI_DRAWINGTYPE_NORMAL,
		FSGUI_DRAWINGTYPE_ACTIVE,
		FSGUI_DRAWINGTYPE_INACTIVE
	};

	static SCHEME scheme;

	static YsColor defBgCol,defFgCol;
	static YsColor defActiveBgCol,defActiveFgCol;
	static YsColor defFrameCol;

	static YsColor defDialogBgCol;

	static YsColor defTabBgCol;
	static YsColor defTabClosedBgCol;
	static YsColor defTabClosedFgCol;

	// used in MODERN scheme
	static double defRoundRadius;

	// for ListBox and DropList
	static YsColor defListBgCol,defListFgCol;
	static YsColor defListActiveBgCol,defListActiveFgCol;

	static class YsFontRenderer *defUnicodeRenderer;
	static class YsFontRenderer *defAsciiRenderer;

	static int defHSpaceUnit,defVSpaceUnit,defHAnnotation,defHScrollBar;

	static int defBitmapWidthLimit,defBitmapHeightLimit;

	static YSRESULT GetUnicodeCharacterBitmapSize(int &wid,int &hei);  // Based on the size of L"X"
	static YSRESULT RenderUnicodeString(YsBitmap &bmp,const char str[],const YsColor &fgCol,const YsColor &bgCol);
	static YSRESULT RenderUnicodeString(YsBitmap &bmp,const wchar_t wStr[],const YsColor &fgCol,const YsColor &bgCol);
	static YSRESULT GetTightUnicodeRenderSize(int &wid,int &hei,const wchar_t wStr[]);

	static YSRESULT GetAsciiCharacterBitmapSize(int &wid,int &hei);  // Based on the size of "X"
	static YSRESULT GetAsciiStringBitmapSize(int &wid,int &hei,const char *str);
	static YSRESULT RenderAsciiString(YsBitmap &bmp,const char str[],const YsColor &fgCol,const YsColor &bgCol);
	static YSRESULT RenderAsciiString(YsBitmap &bmp,const wchar_t str[],const YsColor &fgCol,const YsColor &bgCol);
	static YSRESULT GetTightAsciiRenderSize(int &wid,int &hei,const char str[]);
};

////////////////////////////////////////////////////////////

class FsGuiFontRenderer
{
public:
	virtual YSRESULT RenderString(YsBitmap &bmp,const wchar_t wStr[],const YsColor &fgCol,const YsColor &bgCol) const=0;
	YSRESULT RenderString(YsBitmap &bmp,const char str[],const YsColor &fgCol,const YsColor &bgCol) const;
	virtual YSRESULT GetTightRenderSize(int &wid,int &hei,const wchar_t wStr[]) const=0;
};

class FsGuiAsciiFixedFontRenderer : public FsGuiFontRenderer
{
public:
	virtual int GetFontWidth(void) const=0;
	virtual int GetFontHeight(void) const=0;
	virtual YSRESULT RenderString(YsBitmap &bmp,const wchar_t wStr[],const YsColor &fgCol,const YsColor &bgCol) const=0;
	virtual YSRESULT GetTightRenderSize(int &wid,int &hei,const wchar_t wStr[]) const=0;
};

////////////////////////////////////////////////////////////

/*! This class is a base class for FsGuiCanvas, FsGuiPopUpmenu, and FsGuiDialog.
    */
class FsGuiControlBase : public FsGuiObject
{
private:
	FsGuiControlBase *parent;
	class FsGuiDialog *modalDlg;

public:
	/*! Default constructor. */
	FsGuiControlBase();

	/*! This function sets a pointer to the parent control-base object. */
	void SetParent(FsGuiControlBase *parent);

	/*! This function returns a pointer to the parent control-base object. */
	FsGuiControlBase *GetParent(void) const;

	/*! This function attaches a pointer to the modal dialog. Giving NULL to modalDlg will detach the dialog. */
	void AttachModalDialog(class FsGuiDialog *modalDlg);

	/*! This function detaches the modal dialog.  Same effect as AttachModalDialog(NULL);. */
	void DetachModalDialog(void);

	/*! This function returns a pointer to the modal dialog attached to this control-base object. */
	class FsGuiDialog *GetModalDialog(void) const;

	/*! This function is called when a modal dialog that belongs to this object has been closed. */
    virtual void OnModalDialogClosed(int dialogIdent,class FsGuiDialog *closedModalDialog,int modalDialogEndCode);

	/*! Override this function to implement behavior of the Save button of the View-control dialog. 
	    (A tunnel for 3D-enabled applications.) */
	virtual YSRESULT OnSaveView(const class FsGui3DViewControlDialog *dlg);

	/*! Override this function to implement behavior of the Loae button of the View-control dialog.
	    (A tunnel for 3D-enabled applications.) */
	virtual YSRESULT OnLoadView(class FsGui3DViewControlDialog *dlg);

	/*! The framework (main loop) must call this function regularly. */
	virtual void Interval(void);

	/*! This function must return YSTRUE when a content of the object needs to be re-drawn. */
	virtual YSBOOL NeedRedraw(void) const=0;
	// Memo to myself: The condition for NeedRedraw may also depend on the child objects.  Therefore, it needs to be implemented per class.

	/*! This function sets need-redraw flag. */
	virtual void SetNeedRedraw(YSBOOL needRedraw)=0;
};


/* } */
#endif
