#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <time.h>

#include <ysclass.h>

#include <fssimplewindow.h>
#include "fsguiclipboard.h"

#include <ysbitmap.h>

#include "fsgui.h"
#include "fsguicommondrawing.h"
#include "fsguistatusbar.h"


FsGuiCanvas::FsGuiCanvas()
{
	Initialize();
}

void FsGuiCanvas::Initialize(void)
{
	mainMenu=NULL;
	activeDlg=NULL;
	topStatusBar=NULL;
	bottomStatusBar=NULL;
	dlgArray.Clear();
	prevWid=100;
	prevHei=100;
}

void FsGuiCanvas::SetMainMenu(class FsGuiPopUpMenu *mainMenu)
{
	this->mainMenu=mainMenu;
	if(NULL!=mainMenu)
	{
		mainMenu->SetParent(this);
	}
}

void FsGuiCanvas::SetTopStatusBar(class FsGuiStatusBar *topBar)
{
	topStatusBar=topBar;
}

class FsGuiStatusBar *FsGuiCanvas::GetTopStatusBar(void) const
{
	return topStatusBar;
}

void FsGuiCanvas::SetBottomStatusBar(class FsGuiStatusBar *bottomBar)
{
	bottomStatusBar=bottomBar;
}

class FsGuiStatusBar *FsGuiCanvas::GetBottomStatusBar(void) const
{
	return bottomStatusBar;
}

class FsGuiPopUpMenu *FsGuiCanvas::GetMainMenu(void) const
{
	return mainMenu;
}

YSRESULT FsGuiCanvas::AddDialog(class FsGuiDialog *dlg)
{
	if(NULL!=dlg && YSTRUE!=dlgArray.IsIncluded(dlg))
	{
		dlgArray.Append(dlg);
		dlg->SetParent(this);
		return YSOK;
	}
	return YSERR;
}

YSRESULT FsGuiCanvas::RemoveDialog(class FsGuiDialog *dlg)
{
	for(YSSIZE_T i=0; i<dlgArray.GetN(); ++i)
	{
		if(dlg==dlgArray[i])
		{
			if(dlg==activeDlg)
			{
				if(0<i)
				{
					activeDlg=dlgArray[i-1];
				}
				else if(i<dlgArray.GetN()-1)
				{
					activeDlg=dlgArray[i+1];
				}
				else
				{
					activeDlg=NULL;
				}
			}
			dlgArray.Delete(i);
			SetNeedRedraw(YSTRUE);
			return YSOK;
		}
	}
	return YSERR;
}

void FsGuiCanvas::RemoveDialogAll(void)
{
	if(NULL!=activeDlg)
	{
		activeDlg=NULL;
	}
	dlgArray.Clear();
	SetNeedRedraw(YSTRUE);
}

void FsGuiCanvas::ArrangeDialog(void)
{
	int wid,hei;
	FsGetWindowSize(wid,hei);

	int topLeftX0,topLeftY0;
	int topRightX0,topRightY0;
	int bottomLeftX0,bottomLeftY0;
	int bottomRightX0,bottomRightY0;

	topLeftX0=0;
	topRightX0=wid;
	if(GetMainMenu()!=NULL)
	{
		topLeftY0=GetMainMenu()->GetHeight();
		topRightY0=GetMainMenu()->GetHeight();
	}
	else
	{
		topLeftY0=0;
		topRightY0=0;
	}
	if(NULL!=GetTopStatusBar())
	{
		topLeftY0+=GetTopStatusBar()->GetHeight();
		topRightY0+=GetTopStatusBar()->GetHeight();
	}


	bottomLeftX0=0;
	bottomLeftY0=hei;
	bottomRightX0=wid;
	bottomRightY0=hei;

	if(NULL!=GetBottomStatusBar())
	{
		bottomLeftY0-=GetBottomStatusBar()->GetHeight();
		bottomRightY0-=GetBottomStatusBar()->GetHeight();
	}

	for(int i=0; i<GetNumDialog(); ++i)
	{
		FsGuiDialog *dlg=GetDialog(i);
		switch(dlg->GetArrangeType())
		{
		case FSDIALOG_ARRANGE_TOP_LEFT:
			dlg->Move(topLeftX0,topLeftY0);
			topLeftY0+=dlg->GetHeight();
			break;
		case FSDIALOG_ARRANGE_TOP_RIGHT:
			dlg->Move(topRightX0-dlg->GetWidth(),topRightY0);
			topRightY0+=dlg->GetHeight();
			break;
		case FSDIALOG_ARRANGE_BOTTOM_LEFT:
			dlg->Move(bottomLeftX0,bottomLeftY0-dlg->GetHeight());
			bottomLeftY0-=dlg->GetHeight();
			break;
		case FSDIALOG_ARRANGE_BOTTOM_RIGHT:
			dlg->Move(bottomRightX0-dlg->GetWidth(),bottomRightY0-dlg->GetHeight());
			bottomRightY0-=dlg->GetHeight();
			break;
		}
	}

	SetNeedRedraw(YSTRUE);
}

void FsGuiCanvas::KillDialogFocusAll(void)
{
	for(int i=0; i<GetNumDialog(); ++i)
	{
		GetDialog(i)->SetFocus(NULL);
	}
	FsGuiDialog *activeModalDlg=GetActiveModalDialog();
	if(NULL!=activeModalDlg)
	{
		activeModalDlg->SetFocus(NULL);
	}

	SetNeedRedraw(YSTRUE);
}

int FsGuiCanvas::GetNumDialog(void) const
{
	return dlgArray.GetN();
}

class FsGuiDialog *FsGuiCanvas::GetDialog(int index) const
{
	if(YSTRUE==dlgArray.IsInRange(index))
	{
		return dlgArray[index];
	}
	return NULL;
}

YSBOOL FsGuiCanvas::IsDialogUsed(const FsGuiDialog *dlg) const
{
	for(YSSIZE_T dlgIdx=0; dlgIdx<GetNumDialog(); ++dlgIdx)
	{
		if(dlg==GetDialog(dlgIdx))
		{
			return YSTRUE;
		}
	}

	for(FsGuiDialog *modalDlg=this->GetModalDialog(); NULL!=modalDlg; modalDlg=modalDlg->GetModalDialog())
	{
		if(dlg==modalDlg)
		{
			return YSTRUE;
		}
	}

	return YSFALSE;
}

FsGuiDialog *FsGuiCanvas::GetActiveModalDialog(void) const
{
	if(NULL!=this->GetModalDialog())
	{
		FsGuiDialog *dlg=this->GetModalDialog();
		while(NULL!=dlg->GetModalDialog())
		{
			dlg=dlg->GetModalDialog();
		}
		return dlg;
	}

	for(YSSIZE_T i=0; i<dlgArray.GetN(); ++i)
	{
		FsGuiDialog *dlg=dlgArray[i]->GetModalDialog();
		if(NULL!=dlg)
		{
			while(NULL!=dlg->GetModalDialog())
			{
				dlg=dlg->GetModalDialog();
			}
			return dlg;
		}
	}
	if(NULL!=mainMenu && NULL!=mainMenu->GetModalDialog())
	{
		FsGuiDialog *dlg=mainMenu->GetModalDialog();
		while(NULL!=dlg->GetModalDialog())
		{
			dlg=dlg->GetModalDialog();
		}
		return dlg;
	}

	return NULL;
}

FsGuiDialog *FsGuiCanvas::GetActiveDialog(void) const
{
	return activeDlg;
}

YSRESULT FsGuiCanvas::SetActiveDialog(const class FsGuiDialog *newActiveDlg)
{
	if(NULL==newActiveDlg)
	{
		activeDlg=NULL;
		return YSOK;
	}

	for(int i=0; i<dlgArray.GetN(); ++i)
	{
		if(dlgArray[i]==newActiveDlg)
		{
			activeDlg=dlgArray[i];
			return YSOK;
		}
	}

	return YSERR;
}

void FsGuiCanvas::Interval(void)
{
	// If file-dialog is set, open one and then call
	//     virtual void OnFileSelected(int fileDialogIdent,int nFile,const YsWString fileName[]);
	//
}

YSRESULT FsGuiCanvas::SetMouseState(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my)
{
	YSBOOL eventTaken=YSFALSE;
	int prevMx=this->prevMx;
	int prevMy=this->prevMy;
	YSBOOL prevLb=this->prevLb;
	YSBOOL prevMb=this->prevMb;
	YSBOOL prevRb=this->prevRb;

	this->prevMx=mx;
	this->prevMy=my;
	this->prevLb=lb;
	this->prevMx=mx;
	this->prevRb=rb;



	FsGuiDialog *modalDlg=GetActiveModalDialog();
	if(NULL!=modalDlg)
	{
		return modalDlg->SetMouseState(lb,mb,rb,mx,my);
	}

	if(NULL!=mainMenu && YSOK==mainMenu->SetMouseState(lb,mb,rb,mx,my))
	{
		ForgetPreviousMouseButtonState();
		return YSOK;
	}

	if(NULL!=mainMenu && YSTRUE==mainMenu->IsMouseOnMenu(mx,my))
	{
		eventTaken=YSTRUE;
	}

	for(YSSIZE_T i=dlgArray.GetN()-1; 0<=i; --i)
	{
		auto dlgPtr=dlgArray[i];  // This dialog may be closed inside SetMouseState.  Therefore, the pointer must be saved.
		if(YSOK==dlgPtr->SetMouseState(lb,mb,rb,mx,my))
		{
			if(YSTRUE==dlgArray.IsIncluded(dlgPtr))  // Make sure the dialog was not closed inside SetMouseState.
			{
				activeDlg=dlgPtr;
			}
			return YSOK;
		}
		if(YSTRUE==dlgPtr->IsMouseOnDialog(mx,my))
		{
			return YSOK;
		}
	}


	if(prevLb!=YSTRUE && lb==YSTRUE)
	{
		if(YSOK==OnLButtonDown(lb,mb,rb,mx,my))
		{
			eventTaken=YSTRUE;
		}
	}
	if(prevMb!=YSTRUE && mb==YSTRUE)
	{
		if(YSOK==OnMButtonDown(lb,mb,rb,mx,my))
		{
			eventTaken=YSTRUE;
		}
	}
	if(prevRb!=YSTRUE && rb==YSTRUE)
	{
		if(YSOK==OnRButtonDown(lb,mb,rb,mx,my))
		{
			eventTaken=YSTRUE;
		}
	}

	if(prevMx!=mx || prevMy!=my)
	{
		if(YSOK==OnMouseMove(lb,mb,rb,mx,my,prevMx,prevMy))
		{
			eventTaken=YSTRUE;
		}
	}

	if(prevLb==YSTRUE && lb!=YSTRUE)
	{
		if(YSOK==OnLButtonUp(lb,mb,rb,mx,my))
		{
			eventTaken=YSTRUE;
		}
	}
	if(prevMb==YSTRUE && mb!=YSTRUE)
	{
		if(YSOK==OnMButtonUp(lb,mb,rb,mx,my))
		{
			eventTaken=YSTRUE;
		}
	}
	if(prevRb==YSTRUE && rb!=YSTRUE)
	{
		if(YSOK==OnRButtonUp(lb,mb,rb,mx,my))
		{
			eventTaken=YSTRUE;
		}
	}

	if(YSTRUE==eventTaken)
	{
		return YSOK;
	}


	return YSERR;
}

void FsGuiCanvas::ForgetPreviousMouseButtonState(void)
{
	for(YSSIZE_T i=dlgArray.GetN()-1; 0<=i; --i)
	{
		dlgArray[i]->ForgetPreviousMouseButtonState();
	}
}


YSRESULT FsGuiCanvas::KeyIn(int fsKey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt)
{
	// 2013/09/21
	//   From the requirement of the new polygon editor.
	//     Overrided OnKeyDown must first be tested, otherwise, space key is passed to the view-control dialog.
	//     But, key stroke for the menu must have the priority.  Otherwise, sub-menu cannot be selected by short-cut keys.
	//
	//     However, if FsGuiTextBox is active, all the key strokes must be sent to the dialog, therefore, activeDialog has the priority.
	//
	//     In conclusion, the priority changes depending on the state of the user-interface objects.
	//
	//     Modal dialog always has the top priority.
	//     The main menu takes the key if a sub-menu is open.
	//
	//     The active dialog has the priority over the canvas if MayEatThisKeyStroke returns YSTRUE.
	//     Otherwise, the canvas has the priority over the active dialog.

	FsGuiDialog *modalDlg=GetActiveModalDialog();
	if(NULL!=modalDlg)
	{
		return modalDlg->KeyIn(fsKey,shift,ctrl,alt);;
	}

	if(NULL!=mainMenu && NULL!=mainMenu->GetOpenSubMenu())  // If a sub-menu is open, the main menu takes the key.
	{
		if(NULL!=mainMenu && YSOK==mainMenu->KeyIn(fsKey,shift,ctrl,alt))
		{
			return YSOK;
		}
	}
	else // The main menu has the last priority
	{
		if(NULL!=activeDlg && YSTRUE==activeDlg->MayEatThisKeyStroke(fsKey,shift,ctrl,alt)) // Active dialog has the first priority
		{
			if(NULL!=activeDlg && YSOK==activeDlg->KeyIn(fsKey,shift,ctrl,alt))
			{
				return YSOK;
			}
			if(YSOK==OnKeyDown(fsKey,shift,ctrl,alt))
			{
				return YSOK;
			}
		}
		else // Canvas has the first priority
		{
			if(YSOK==OnKeyDown(fsKey,shift,ctrl,alt))
			{
				return YSOK;
			}
			if(NULL!=activeDlg && YSOK==activeDlg->KeyIn(fsKey,shift,ctrl,alt))
			{
				return YSOK;
			}
		}

		if(NULL!=mainMenu && YSOK==mainMenu->KeyIn(fsKey,shift,ctrl,alt))
		{
			return YSOK;
		}
	}

	return YSERR;
}

YSRESULT FsGuiCanvas::SetWindowSize(int newWid,int newHei,YSBOOL autoArrangeDialog)
{
	if(newWid!=prevWid || newHei!=prevHei)
	{
		prevWid=newWid;
		prevHei=newHei;
		if(YSTRUE==autoArrangeDialog)
		{
			ArrangeDialog();
		}
		return YSOK;
	}
	return YSERR;
	}

YSRESULT FsGuiCanvas::CharIn(int c)
{
	FsGuiDialog *modalDlg=GetActiveModalDialog();
	if(NULL!=modalDlg)
	{
		return modalDlg->CharIn(c);
	}

	if(NULL!=activeDlg && YSOK==activeDlg->CharIn(c))
	{
		return YSOK;
	}

	if(YSOK==OnKeyChar(c))
	{
		return YSOK;
	}

	return YSERR;
}

void FsGuiCanvas::Show(void) const
{
	const FsGuiDialog *modalDlg=GetModalDialog();
	int wid,hei;
	FsGetWindowSize(wid,hei);

	if(NULL!=topStatusBar)
	{
		const int y0=(NULL==mainMenu ? 0 : mainMenu->GetHeight());
		topStatusBar->SetWidth(wid);
		topStatusBar->Show(y0);
	}
	if(NULL!=bottomStatusBar)
	{
		const int y0=hei-bottomStatusBar->GetHeight();
		bottomStatusBar->SetWidth(wid);
		bottomStatusBar->Show(y0);
	}

	for(YSSIZE_T i=0; i<dlgArray.GetN(); ++i)
	{
		dlgArray[i]->Show(modalDlg);
	}

	if(NULL!=mainMenu)
	{
		mainMenu->SetWidth(wid);
		mainMenu->Show(modalDlg);
	}
	if(NULL!=modalDlg)
	{
		FsGuiCommonDrawing::DrawMask(FsGuiObject::defFgCol,FsGuiObject::defBgCol,0.0,0.0,(double)wid,(double)hei);
		modalDlg->Show();
	}
}

YSBOOL FsGuiCanvas::NeedRedraw(void) const
{
	if(NULL!=GetModalDialog() && YSTRUE==GetModalDialog()->NeedRedraw())
	{
		return YSTRUE;
	}
	if(NULL!=mainMenu && YSTRUE==mainMenu->NeedRedraw())
	{
		return YSTRUE;
	}
	if(NULL!=GetTopStatusBar() && YSTRUE==GetTopStatusBar()->NeedRedraw())
	{
		return YSTRUE;
	}
	if(NULL!=GetBottomStatusBar() && YSTRUE==GetBottomStatusBar()->NeedRedraw())
	{
		return YSTRUE;
	}
	for(YSSIZE_T i=0; i<dlgArray.GetN(); ++i)
	{
		if(YSTRUE==dlgArray[i]->NeedRedraw())
		{
			return YSTRUE;
		}
	}
	return needRedraw;
}

void FsGuiCanvas::SetNeedRedraw(YSBOOL needRedraw)
{
	this->needRedraw=needRedraw;
}


YSRESULT FsGuiCanvas::OnKeyDown(int,YSBOOL,YSBOOL,YSBOOL)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnKeyChar(int)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnLButtonDown(YSBOOL,YSBOOL,YSBOOL,int,int)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnMButtonDown(YSBOOL,YSBOOL,YSBOOL,int,int)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnRButtonDown(YSBOOL,YSBOOL,YSBOOL,int,int)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnMouseMove(YSBOOL,YSBOOL,YSBOOL,int,int,int,int)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnLButtonUp(YSBOOL,YSBOOL,YSBOOL,int,int)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnMButtonUp(YSBOOL,YSBOOL,YSBOOL,int,int)
{
	return YSERR;
}

YSRESULT FsGuiCanvas::OnRButtonUp(YSBOOL,YSBOOL,YSBOOL,int,int)
{
	return YSERR;
}

