/* ////////////////////////////////////////////////////////////

  File Name: ysshellext_localop.h
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#ifndef YSSHELLEXT_LOCALOP_IS_INCLUDED
#define YSSHELLEXT_LOCALOP_IS_INCLUDED
/* { */

#include <ysshellext.h>


class YsShell_LocalOperation
{
public:
	class PolygonChange
	{
	public:
		YsShellPolygonHandle plHd;
		YsArray <YsShellVertexHandle,4> plVtHd;
	};
	class ConstEdgeChange
	{
	public:
		YsShellExt::ConstEdgeHandle ceHd;
		YsArray <YsShellVertexHandle> ceVtHd;
		YSBOOL isLoop;
	};

	class EdgeSplit
	{
	public:
		YsShellVertexHandle edVtHd[2];
		YsVec3 pos;
		YsShellVertexHandle createdVtHd;

		inline void CleanUp(void)
		{
			edVtHd[0]=NULL;
			edVtHd[1]=NULL;
			createdVtHd=NULL;
		}
	};

	class PolygonSplit
	{
	public:
		YsShellPolygonHandle plHd;
		YsSegmentedArray <YsArray <YsShellVertexHandle,4>,4> plVtHdArray;
		YsSegmentedArray <YsArray <YsVec2,4>,4> prjVtPosArray;
	};
	class ConstEdgeSplit
	{
	public:
		YsShellExt::ConstEdgeHandle ceHd;
		YsSegmentedArray <YsArray <YsShellVertexHandle>,4> ceVtHdArray;
	};
	class FaceGroupSplit
	{
	public:
		YsShellExt::FaceGroupHandle fgHd;
		YsSegmentedArray <YsArray <YsShellPolygonHandle>,4> fgPlHdArray;

		void CleanUp(void);
	};
};


class YsShell_SwapInfo : public YsShell_LocalOperation
{
public:
	YsShellVertexHandle orgDiagonal[2],newDiagonal[2],quadVtHd[4];
	YsShellVertexHandle newTriVtHd[2][3];
	YsShellPolygonHandle triPlHd[2];

	YsShell_SwapInfo();
	void CleanUp(void);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellVertexHandle edVtHd[2]);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle edPlHd[2]);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle edPlHd0,const YsShellPolygonHandle edPlHd1);
};


class YsShell_MergeInfo : public YsShell_LocalOperation
{
public:
	YsShellPolygonHandle plHd[2];
	YsArray <YsShellVertexHandle> newPlVtHd;
	YSBOOL duplicateWarning; // YSTRUE -> New polygon may be using a same vertex more than once.

	YsShell_MergeInfo();
	void CleanUp(void);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle plHd[2]);
	YSRESULT MakeInfo(const YsShell &shl,const YsShellPolygonHandle plHd0,const YsShellPolygonHandle plHd1);
};

class YsShell_CollapseInfo : public YsShell_LocalOperation
{
public:
	YsSegmentedArray <PolygonChange,4> newPlg;
	YsSegmentedArray <ConstEdgeChange,4> newConstEdge;
	YsArray <YsShellPolygonHandle> plHdToDel;
	YsArray <YsShellExt::ConstEdgeHandle> ceHdToDel;
	YsShellVertexHandle vtHdToDel,vtHdToSurvive;

	YsShell_CollapseInfo();
	void CleanUp(void);
	YSRESULT MakeInfo(const YsShellExt &shl,const YsShellVertexHandle vtHd[2]);
	YSRESULT MakeInfo(const YsShellExt &shl,const YsShellVertexHandle vtHdToDel,const YsShellVertexHandle vtHdToSurvive);
};



/*! This function inserts a vertex on an edge of the vertex array. */
template <class VertexType,const int N>
YSRESULT YsShellExt_InsertVertexOnEdgeOfVertexArray(
    YsArray <VertexType,N> &vtHdArray,YSBOOL isLoop,const VertexType edVtHd[2],VertexType vtHd)
{
	if(edVtHd[0]==vtHd || edVtHd[1]==vtHd || 2>vtHdArray.GetN())
	{
		return YSERR;
	}

	YSRESULT found=YSERR;
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN()-1; ++idx)
	{
		if(YSTRUE==YsSameEdge(vtHdArray[idx],vtHdArray[idx+1],edVtHd[0],edVtHd[1]))
		{
			vtHdArray.Insert(idx+1,vtHd);
			found=YSOK;
		}
	}

	if(YSTRUE==isLoop && YSTRUE==YsSameEdge(vtHdArray.GetTop(),vtHdArray.GetEnd(),edVtHd[0],edVtHd[1]))
	{
		vtHdArray.Append(vtHd);
		found=YSOK;
	}

	return found;
}


/*! This function splits a loop into two loops at index idx1 and idx2.  
    The original (input) loop is loopVtHd0.  
    After split, the loop becomes two loops and will be stored in loopVtHd0 and loopVtHd1. 

	If no error, this function returns YSOK.

    Absolute value of idx0-idx1 must be at least 2, or this function returns YSERR.

    ****!  I'm sure I have written an exactly same function before.  I cannot just find it.
    */
template <class VertexType,const int N>
YSRESULT YsShellExt_SplitLoop(YsArray <VertexType,N> &loopVtHd0,YsArray <VertexType,N> &loopVtHd1,YSSIZE_T idx0,YSSIZE_T idx1)
{
	if(YSTRUE==loopVtHd0.IsInRange(idx0) && YSTRUE==loopVtHd0.IsInRange(idx1))
	{
		if(idx1<idx0)
		{
			YsSwapSomething (idx1,idx0);
		}

		if(2>idx1-idx0 || 2>loopVtHd0.GetN()-(idx1-idx0))
		{
			return YSERR;
		}

		loopVtHd1.Set(idx1-idx0+1,loopVtHd0.GetArray()+idx0);

		const YSSIZE_T toDel=idx1-idx0-1;
		loopVtHd0.Delete(idx0+1,toDel);

		return YSOK;
	}
	return YSERR;
}


/*! This function splits an open chain, chain0, at index of idx0 into two separate chains. 

	The first of the two will be returned in chain0, and the rest in chain1.
    */
template <class VertexType,const int N>
YSRESULT YsShellExt_SplitOpenChain(YsArray <VertexType,N> &chain0,YsArray <VertexType,N> &chain1,YSSIZE_T idx0)
{
	if(1<=idx0 && idx0<chain0.GetN()-1)
	{
		chain1.Set(chain0.GetN()-idx0,chain0.GetArray()+idx0);
		chain0.Resize(idx0+1);
		return YSOK;
	}
	return YSERR;
}

/* } */
#endif
