/* ////////////////////////////////////////////////////////////

  File Name: ysshellext_orientationutil.h
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#ifndef YSSHELLEXT_ORIENTATIONUTIL_IS_INCLUDED
#define YSSHELLEXT_ORIENTATIONUTIL_IS_INCLUDED
/* { */

#include <ysshellextedit.h>

class YsShellExt_OrientationUtil
{
friend class YsShellExtEdit_OrientationUtil;

public:
	class PolygonNormalPair
	{
	public:
		YsShellPolygonHandle plHd;
		YsVec3 nom;
	};

private:
	const YsShellExt *shl;
	YsShellPolygonStore visitedPlHd;
	YsArray <YsShellPolygonHandle> plHdNeedFlipArray;
	YsArray <PolygonNormalPair> plHdNeedNormalChangeArray;

	YSBOOL abortFlag;
	YSBOOL fixOrientetionFromReliablePolygonEnded;
	YSBOOL fixOrientationOfClosedSolidEnded;
	YSBOOL normalRecalculationEnded;

	YSBOOL staticCreate;

public:
	YsShellExt_OrientationUtil();
	~YsShellExt_OrientationUtil();

	/*! This sets abort flag.  In multi-threaded environment, this will terminate a process running on this object. */
	void Abort(void);

	/*! This function returns YSTRUE if the process 

	/*! Cleans up the internal data structure. */
	void CleanUp(void);

	/*! This function assumes that seedPlHd's orientation is reliable, and makes a list of polygons that need to be flipped starting from a seedPlHd. 
	    This function will not clear the list of polygons if it already exists. */
	YSRESULT FixOrientationFromReliablePolygon(const YsShellExt &shl,YsShellPolygonHandle seedPlHd,YSBOOL seedPlgNeedFlip);

	/*! This function assumes that seedPlHd's orientation is reliable, and makes a list of polygons that need to be flipped starting from a seedPlHd. 
	    This function will not clear the list of polygons if it already exists. 
	    The number of polygons covered in this function will be returned in nPlgCovered. */
	YSRESULT FixOrientationFromReliablePolygon(YSSIZE_T &nPlgCovered,const YsShellExt &shl,YsShellPolygonHandle seedPlHd,YSBOOL seedPlgNeedFlip);

	/*! This function makes a list of polygons that need to be flipped.  
	    Call CleanUp function before calling this function. */
	YSRESULT FixOrientationOfClosedSolid(const YsShellExt &shl);

	/*! This function re-calculate normal based on the polygon orientation and
	    makes list of polygon-normal array that stores updated normal. */
	YSRESULT RecalculateNormalFromCurrentOrientation(const YsShell &shl);

	/*! This function re-calculate normal of the polygons plHd based on the polygon orientation and
	    makes list of polygon-normal array that stores updated normal. */
	YSRESULT RecalculateNormalFromCurrentOrientation(const YsShell &shl,YsShellPolygonHandle plHd);

	/*! This function re-calculate normal of the polygons plHd[0] ... plHd[nPl-1] based on the polygon orientation and
	    makes list of polygon-normal array that stores updated normal. */
	YSRESULT RecalculateNormalFromCurrentOrientation(const YsShell &shl,YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[]);

	/*! This function re-calculate normal of the polygons plHd[0] ... plHd[nPl-1] based on the polygon orientation and
	    makes list of polygon-normal array that stores updated normal. */
	template <const int N>
	YSRESULT RecalculateNormalFromCurrentOrientation(const YsShell &shl,const YsArray <YsShellPolygonHandle,N> &plHdArray);

	/*! This function returns a pointer to the array of polygons that needs to be flipped. */
	const YsArray <YsShellPolygonHandle> &GetPolygonNeedFlip(void) const;

	/*! This function returns polygon-normal pairs of the polygons that the normal vector needs to be reset. */
	const YsArray <PolygonNormalPair> &GetPolygonNormalPair(void) const;

	/*! This function inverts all polygons in YsShellExt */
	static void InvertPolygonAll(YsShellExt &shl);
};

template <const int N>
YSRESULT YsShellExt_OrientationUtil::RecalculateNormalFromCurrentOrientation(const YsShell &shl,const YsArray <YsShellPolygonHandle,N> &plHdArray)
{
	return RecalculateNormalFromCurrentOrientation(shl,plHdArray.GetN(),plHdArray);
}

/* } */
#endif
