/* ////////////////////////////////////////////////////////////

  File Name: ysshellext_trackingutil.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "ysshellext_trackingutil.h"


YsShellExt_TrackingUtil::SearchCondition::SearchCondition()
{
	Initialize();
}

void YsShellExt_TrackingUtil::SearchCondition::Initialize(void)
{
	withinThesePlHd=NULL;
	excludeThesePlHd=NULL;
	onlyInThisFgHd=NULL;
}

YSRESULT YsShellExt_TrackingUtil::SearchCondition::TestPolygon(const YsShellExt &shl,YsShellPolygonHandle plHd) const
{
	if(NULL!=withinThesePlHd && YSTRUE!=withinThesePlHd->IsIncluded(plHd))
	{
		return YSERR;
	}
	if(NULL!=excludeThesePlHd && YSTRUE==excludeThesePlHd->IsIncluded(plHd))
	{
		return YSERR;
	}
	if(NULL!=onlyInThisFgHd)
	{
		auto plFgHd=shl.FindFaceGroupFromPolygon(plHd);
		if(plFgHd!=onlyInThisFgHd)
		{
			return YSERR;
		}
	}
	return YSOK;
}

/* virtual */ YSRESULT YsShellExt_TrackingUtil::SearchCondition::TestEdge(const YsShellExt &,YsShellVertexHandle,YsShellVertexHandle) const
{
	return YSOK;
}

////////////////////////////////////////////////////////////


/*static*/ YsArray <YsShellVertexHandle> YsShellExt_TrackingUtil::TrackVertexConnection(
	const YsShell &shl,
	YsShellVertexHandle fromVtHd,YsShellVertexHandle toVtHd,
	const YsFixedLengthToMultiHashTable <YsShellVertexHandle,1,2> &connection)
{
	YsArray <YsShellVertexHandle> trackedPath;

	trackedPath.Append(fromVtHd);

	YsShellVertexHandle prvVtHd=fromVtHd,curVtHd=toVtHd;
	while(NULL!=curVtHd)
	{
		trackedPath.Append(curVtHd);

		if(fromVtHd==curVtHd) // Closed a loop?
		{
			break;
		}

		unsigned curVtKey[1]=
		{
			shl.GetSearchKey(curVtHd)
		};

		YsShellVertexHandle nxtVtHd=NULL;

		YSSIZE_T nConn;
		const YsShellVertexHandle *connVtHd;
		if(YSOK==connection.FindElement(nConn,connVtHd,1,curVtKey) && 2>=nConn)
		{
			if(1==nConn && connVtHd[0]!=prvVtHd)
			{
				nxtVtHd=connVtHd[0];
			}
			else if(2==nConn)
			{
				if(prvVtHd!=connVtHd[1])
				{
					nxtVtHd=connVtHd[1];
				}
				else if(prvVtHd!=connVtHd[0])
				{
					nxtVtHd=connVtHd[0];
				}
			}
		}

		prvVtHd=curVtHd;
		curVtHd=nxtVtHd;
	}

	return trackedPath;
}

/*static*/ YsArray <YsShellPolygonHandle> YsShellExt_TrackingUtil::TrackPolygonOneSideOfVertexSequence(
    const YsShellExt &shl,
    YSSIZE_T nVt,const YsShellVertexHandle vtHdArray[],
	YSSIZE_T nFlagPl,const YsShellPolygonHandle flagPlHdArray[])
{
	YsArray <YsShellPolygonHandle> empty;

	if(YSTRUE!=shl.IsSearchEnabled())
	{
		YsPrintf("%s %d\n",__FUNCTION__,__LINE__);
		YsPrintf("  This function requires a search table.\n");

		return empty;
	}

	if(2>nVt)
	{
		return empty;
	}

	YsTwoKeyStore edgePiece;
	for(YSSIZE_T idx=0; idx<nVt-1; ++idx)
	{
		edgePiece.AddKey(shl.GetSearchKey(vtHdArray[idx]),shl.GetSearchKey(vtHdArray[idx+1]));
	}

	YsShellVertexHandle tipVtHd[2]={NULL,NULL};
	if(vtHdArray[0]!=vtHdArray[nVt-1])
	{
		// Means an open vertex sequence
		tipVtHd[0]=vtHdArray[0];
		tipVtHd[1]=vtHdArray[nVt-1];
	}

	YsShellVertexStore vtxStore((const YsShell &)shl);
	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		vtxStore.AddVertex(vtHdArray[idx]);
	}

	YsArray <YsShellPolygonHandle> seedPlHdArray,nonSeedPlHdArray;
	for(YSSIZE_T plIdx=0; plIdx<nFlagPl; ++plIdx)
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetPolygon(plVtHd,flagPlHdArray[plIdx]);
		for(YSSIZE_T vtIdx=0; vtIdx<plVtHd.GetN(); ++vtIdx)
		{
			if(YSTRUE==edgePiece.IsIncluded(
				shl.GetSearchKey(plVtHd[vtIdx]),
				shl.GetSearchKey(plVtHd.GetCyclic(vtIdx+1))))
			{
				seedPlHdArray.Append(flagPlHdArray[plIdx]);
				goto NEXTFLAGPLG;
			}
		}
		nonSeedPlHdArray.Append(flagPlHdArray[plIdx]);

	NEXTFLAGPLG:
		;
	}

	if(0==seedPlHdArray.GetN())
	{
printf("%s %d\n",__FUNCTION__,__LINE__);
		return empty;
	}


	YsArray <YsShellPolygonHandle> plHdArray;
	plHdArray.Append(nonSeedPlHdArray);
	plHdArray.Append(seedPlHdArray);

	YsShellPolygonStore visited((const YsShell &)shl);
	visited.AddPolygon(plHdArray);


	if(2<=nFlagPl)
	{
		// Polygons need to be edge-connected.
		YsShellExt_TrackingUtil::SearchCondition cond;
		cond.withinThesePlHd=&visited;
		YsArray <YsShellPolygonHandle> verify=YsShellExt_TrackingUtil::MakeEdgeConnectedPolygonGroup(shl,plHdArray[0],&cond);
		if(verify.GetN()!=nFlagPl)
		{
printf("%s %d\n",__FUNCTION__,__LINE__);
			return empty;
		}
	}

printf("%s %d\n",__FUNCTION__,__LINE__);
	
	for(YSSIZE_T ptr=nonSeedPlHdArray.GetN(); ptr<plHdArray.GetN(); ++ptr)
	{
		YsShellPolygonHandle plHd=plHdArray[ptr];

		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetPolygon(plVtHd,plHd);

		for(YSSIZE_T vtIdx=0; vtIdx<plVtHd.GetN(); ++vtIdx)
		{
			const YsShellVertexHandle edVtHd[2]=
			{
				plVtHd[vtIdx],
				plVtHd.GetCyclic(vtIdx+1)
			};

			if(edVtHd[0]!=tipVtHd[0] &&
			   edVtHd[1]!=tipVtHd[0] &&
			   edVtHd[0]!=tipVtHd[1] &&
			   edVtHd[1]!=tipVtHd[1] &&
			   YSTRUE!=edgePiece.IsIncluded(shl.GetSearchKey(edVtHd[0]),shl.GetSearchKey(edVtHd[1])) &&
			   (YSTRUE==vtxStore.IsIncluded(edVtHd[0]) || YSTRUE==vtxStore.IsIncluded(edVtHd[1])))
			{
				YsShellPolygonHandle neiPlHd=shl.GetNeighborPolygon(plHd,edVtHd[0],edVtHd[1]);
				if(NULL!=neiPlHd && YSTRUE!=visited.IsIncluded(neiPlHd))
				{
					visited.AddPolygon(neiPlHd);
					plHdArray.Append(neiPlHd);
				}
			}
		}
	}

printf("%s %d\n",__FUNCTION__,__LINE__);

	return plHdArray;
}

/*static*/ YsArray <YsShellPolygonHandle> YsShellExt_TrackingUtil::MakeEdgeConnectedPolygonGroup(
	    const YsShellExt &shl,YsShellPolygonHandle seedPlHd,SearchCondition *searchCondition)
{
	YsArray <YsShellPolygonHandle> plHdArray;
	YsShellPolygonStore visited((const YsShell &)shl);

	if(YSTRUE!=shl.IsSearchEnabled())
	{
		YsPrintf("%s %d\n",__FUNCTION__,__LINE__);
		YsPrintf("  This function requires a search table.\n");
		return plHdArray;
	}

	plHdArray.Append(seedPlHd);
	visited.AddPolygon(seedPlHd);
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		const int nPl=shl.GetNumVertexOfPolygon(plHdArray[idx]);
		auto plVtHd=shl.GetPolygonVertex(plHdArray[idx]);
		for(int edIdx=0; edIdx<nPl; ++edIdx)
		{
			YsShellPolygonHandle neiPlHd=shl.GetNeighborPolygon(plHdArray[idx],edIdx);

			if(NULL!=searchCondition && YSOK!=searchCondition->TestEdge(shl,plVtHd[edIdx],plVtHd.GetCyclic(edIdx+1)))
			{
				continue;
			}

			if(NULL!=neiPlHd && YSTRUE!=visited.IsIncluded(neiPlHd))
			{
				visited.AddPolygon(neiPlHd);
				if(NULL!=searchCondition && YSOK!=searchCondition->TestPolygon(shl,neiPlHd))
				{
					continue;  // Go to next neighbor
				}
				plHdArray.Append(neiPlHd);
			}
		}
	}

	return plHdArray;
}

/*! This function returns a sequence of vertices by connecting specified constraint edges. */
/*static*/ YsArray <YsShellVertexHandle> YsShellExt_TrackingUtil::MakeVertexSequenceFromMultiConstEdge(
    const YsShellExt &shl,YSSIZE_T nCe,const YsShellExt::ConstEdgeHandle ceHdArray[])
{
	if(1==nCe)
	{
		YsArray <YsShellVertexHandle> ceVtHd;
		YSBOOL isLoop;
		shl.GetConstEdge(ceVtHd,isLoop,ceHdArray[0]);
		if(1<ceVtHd.GetN() && YSTRUE==isLoop && ceVtHd[0]!=ceVtHd.Last())
		{
			ceVtHd.Append(ceVtHd[0]);
		}
		return ceVtHd;
	}

	YsArray <YsShellVertexHandle> vtSeq;

	YsKeyStore selCeKey;
	for(YSSIZE_T idx=1; idx<nCe; ++idx)
	{
		selCeKey.AddKey(shl.GetSearchKey(ceHdArray[idx]));

		YsArray <YsShellVertexHandle> ceVtHd;
		YSBOOL isLoop;
		shl.GetConstEdge(ceVtHd,isLoop,ceHdArray[idx]);
		if(YSTRUE==isLoop || 1>ceVtHd.GetN())
		{
			vtSeq.CleanUp();
			return vtSeq;
		}
	}

	YSBOOL isLoop;
	shl.GetConstEdge(vtSeq,isLoop,ceHdArray[0]);

	if(YSTRUE==isLoop || 2>vtSeq.GetN())
	{
		vtSeq.CleanUp();
		return vtSeq;
	}

	for(YSSIZE_T nUsed=1; nUsed<nCe; ++nUsed)
	{
		YsArray <YsShellExt::ConstEdgeHandle> vtCeHd;
		shl.FindConstEdgeFromVertex(vtCeHd,vtSeq.Last());
		for(auto neiCeHd : vtCeHd)
		{
			if(YSTRUE==selCeKey.IsIncluded(shl.GetSearchKey(neiCeHd)))
			{
				YsArray <YsShellVertexHandle> neiCeVtHd;
				YSBOOL isLoop;
				shl.GetConstEdge(neiCeVtHd,isLoop,neiCeHd);
				if(neiCeVtHd[0]==vtSeq.Last())
				{
					vtSeq.DeleteLast();
					vtSeq.Append(neiCeVtHd);
					selCeKey.DeleteKey(shl.GetSearchKey(neiCeHd));
					goto NEXT;
				}
				else if(neiCeVtHd.Last()==vtSeq.Last())
				{
					neiCeVtHd.Invert();
					vtSeq.DeleteLast();
					vtSeq.Append(neiCeVtHd);
					selCeKey.DeleteKey(shl.GetSearchKey(neiCeHd));
					goto NEXT;
				}
			}
		}
		shl.FindConstEdgeFromVertex(vtCeHd,vtSeq[0]);
		for(auto neiCeHd : vtCeHd)
		{
			if(YSTRUE==selCeKey.IsIncluded(shl.GetSearchKey(neiCeHd)))
			{
				YsArray <YsShellVertexHandle> neiCeVtHd;
				YSBOOL isLoop;
				shl.GetConstEdge(neiCeVtHd,isLoop,neiCeHd);
				if(neiCeVtHd[0]==vtSeq[0])
				{
					vtSeq.Invert();
					vtSeq.DeleteLast();
					vtSeq.Append(neiCeVtHd);
					selCeKey.DeleteKey(shl.GetSearchKey(neiCeHd));
					goto NEXT;
				}
				else if(neiCeVtHd.Last()==vtSeq[0])
				{
					neiCeVtHd.Invert();
					vtSeq.Invert();
					vtSeq.DeleteLast();
					vtSeq.Append(neiCeVtHd);
					selCeKey.DeleteKey(shl.GetSearchKey(neiCeHd));
					goto NEXT;
				}
			}
		}

		// Didn't find any?
		vtSeq.CleanUp();
		return vtSeq;

	NEXT:
		;
	}
	return vtSeq;
}

