#ifndef YSTHREAD_IS_INCLUDED
#define YSTHREAD_IS_INCLUDED
/* { */

#include <ysclass.h>
#include <ystask.h>

#ifdef _WIN32
	#include <thread>
	#include <mutex>
#else
	#include <pthread.h>
	#include <unistd.h>
#endif


/*! 

Apparently, C++11's std::thread::hardware_concurrency has a bug.  Not in implementation, but in the specification.
It must not be allowed to happen, but std::thread::hardware_concurrency can return zero, as in the C++11 spec.
I have confirmed that std::thread::hardware_concurrency can return zero with g++.
It is a critical flaw in design.  Very sad  :-P
Nonetheless, C++11's standard threading library takes care of most of the multi-threading necessities.
Therefore, this class is almost deprecated.  Please do not use except YsThreadController::GetNumCpu.

 */
class YsMutex
{
public:
#ifdef _WIN32
	std::mutex mutex;
#else
	pthread_mutex_t mutex;
#endif

	YsMutex();
	~YsMutex();

	void Lock(void);
	void Unlock(void);
	YSRESULT LockNoWait(void);
};



class YsThreadController
{
public:
	static unsigned int GetNumCPU(void);
	static YSRESULT RunParallel(int nTask,YsTask *const task[]);

	template <const int N>
	static YSRESULT RunParallel(const YsArray <YsTask *> &taskArray);
};

template <const int N>
YSRESULT YsThreadController::RunParallel(const YsArray <YsTask *> &taskArray)
{
	return RunParallel(taskArray.GetN(),taskArray);
}

/* } */
#endif
